/***
 ***  lp_io.c

   buffered parallel-port I/O routines. These lie on top of the
   "asm IN/OUT" analogues in io_ports.c. For convenience, they define
   "bit indicators" which pack the register of a bit, its offset
   in the byte, and whether it's inverted or not into an int.

   There are two pieces of code in this file. The first part does simple
   buffered bitwise I/O:

       lpb_write(lpb, bit, val)
       val = lpb_test(lpb, bit)
         write to and read the bit indicated by 'bit' in the buffer pointed
         to by lpb.

       lpb_flush(lpb)         writes out the data and control bytes
       lpb_refresh(lpb)       reads in the status byte

   The second part parses a (very simple) configuration file format
   to fill in an array of bit indicators. This is lpb_cfg_read(),
   which takes a filename, a pointer to an array of (struct lp_name)s,
   and the length of the array, and tries to fill in the array from the
   data file. There is a companion function, lpb_dump_names, which
   takes an array or structs and its length, and dumps them to stdout
   for debugging.
 ***
 ***/


#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h> /* for strtoul() */

#include "io_ports.h"
#include "prog84.h"

#ifdef __MSDOS__
#include "dosutil.h"
#endif



/* This turns on debugging output from lpb_cfg_read */
/* #define LPBDEBUG  */

void lpb_write(struct lp_io *lpb, int bit, int val)
{
    int mask = 1 << (bit & BI_OFFSET);
    int idx= (bit & BI_TYPE) >> BI_TYPE_SHIFT;
    if (bit & BI_INVERSE)
	val = !val;
    if (val)
	lpb->wanted[ idx ] |= mask;
    else
	lpb->wanted[ idx ] &= ~mask;
}

int lpb_test(struct lp_io *lpb, int bit)
{
    int mask = 1 << (bit & BI_OFFSET);
    if (lpb->last[ (bit & BI_TYPE) >> BI_TYPE_SHIFT ] & mask)
	return (bit & BI_INVERSE)? 0 : 1;
    else
	return (bit & BI_INVERSE)? 1 : 0;
}

#if 0
void lpb_flush(struct lp_io *lpb)
{
    int i, m=1;
    for (i=0; i<7; i++) /* gratuitous loop construct */ {
	if((lpb->wanted[i] != lpb->last[i])) {
	    out_byte(lpb->base + i, (unsigned char)(lpb->wanted[i]));
	    /* printf(".\b"); fflush(stdout); */
	    lpb->last[i] = lpb->wanted[i];
	}
	m += m; // shift left 1
    }
}
#endif

/* New lpb_flush() / lpb_refresh()
   lpb->ports_in_use controls which ports are used
   thru84/linux only seems to work with "-O6" optimation here.
*/

void lpb_flush(struct lp_io *lpb)
{
    int i, m=1;
    for (i=0; i<7; i++) /* gratuitous loop construct */ {
	if((lpb->ports_in_use & m) && (lpb->wanted[i] != lpb->last[i])) {
	    out_byte(lpb->base + i, (unsigned char)(lpb->wanted[i]));
	    /* printf(".\b"); fflush(stdout); */
	    lpb->last[i] = lpb->wanted[i];
	}
	m += m; // shift left 1
    }
}

#if 0
void lpb_refresh(struct lp_io *lpb) {
    /* printf(".\b"); fflush(stdout); */
	lpb->last[1] = in_byte(lpb->base + 1);
	lpb->last[6] = in_byte(lpb->base + 6);
}
#endif

void lpb_refresh(struct lp_io *lpb) {
    /* printf(".\b"); fflush(stdout); */
    //fprintf(stderr,"%02x ",lpb->ports_in_use);
    //	SLEEP100ns(1);
    if (lpb->ports_in_use & 0x02) {
	lpb->last[1] = in_byte(lpb->base + 1);
    }
    if (lpb->ports_in_use & 0x40) {
	lpb->last[6] = in_byte(lpb->base + 6);
    }
}

/***
 *** lpb_cfg_read() and helpers
 ***
 *** NOTE that any entry in the names vector with with the name "base"
 *** is treated specially; it is read as an integer (with C syntax for
 *** octal & hex) instead of being parsed as a bit spec
 ***
 *** Also note that the parseBit() routine is very simple minded right
 *** now and should probably be extended a little bit
 ***/

#define MAXLINE 512

static int parseBit(char *);


int
lpb_cfg_read(const char *filename, int namec, struct lp_name *namev)
{
    FILE *fp;
    char buf[MAXLINE];
    char *line, *cp;
    int wordlen;
    enum { wtNone, wtNumber, wtName, wtType } wordtype;
    int word;

    /* init */
    lpb.ports_in_use=0x05; // portbase +0 always used ?
    fp = fopen(filename, "r");
    if (!fp) {
	perror(filename);
	return -1;
    }

    while(fgets(buf, MAXLINE, fp)) {
#ifdef LPBDEBUG
	fprintf(stderr,"<%s>\n", buf); 
#endif

	/* Trim leading whitespace. */
	for (cp = buf; *cp && isspace(*cp); cp++)
	    ;
	line = cp;

	/* Remove comments. */
	if ( (cp=strchr(line, '#'))!=0 )
	    *cp = (char)0;

	/* Skip blank lines. */
	if (!*line)
	    continue;

	/* Extract the first word of the line. */
	for (cp = line; *cp; cp++)
	    if(isspace(*cp) || (*cp == '=') || (*cp == ':'))
		break;

	wordlen = cp - line;

#ifdef LPBDEBUG
	fprintf(stderr,"[key <%.*s>]\n", wordlen, line); 
#endif

	/* Skip past the separator, if any */
	while(*cp && isspace(*cp))
	    cp++;
	if ((*cp == '=') || (*cp == ':'))
	    cp++;

	/* Find out what kind of word this is. */
	wordtype = wtNone;

	if (!wordtype)
	    for (word = 0; word < namec; word ++)
		if (wordlen == strlen(namev[word].name) &&
		    !strncmp(line, namev[word].name, wordlen)) {
		    wordtype = wtName;
#ifdef LPBDEBUG
		    fprintf(stderr,"[index %d]\n", word); 
#endif
		    break;
		}

	if (wordtype == wtName && /* wordlen == 4 && :: <- old stuff */
		namev[word].lptype==lpc_number)
/*		! ( strncmp(line, "base", wordlen) && 
		    strncmp(line, "loop", wordlen)))  :: <- old stuff */
	    wordtype = wtNumber;

	else if (wordtype == wtName && (namev[word].lptype&lpc_type))
	    wordtype = wtType;

#ifdef LPBDEBUG
	fprintf(stderr,"[wordtype = %d]\n", wordtype); 
#endif

	/* Parse the word. */
	switch (wordtype) {
	case wtNone:
	    fprintf(stderr, "%s: don't know what to do with \"%.*s\"\n",
		filename, wordlen, line);
	    break;
	case wtName:
	    *(namev[word].value) = parseBit(cp);
	    break;
	case wtNumber:
	    *(namev[word].value) = (int) strtoul(cp, (char **)NULL, 0);
	    break;
	case wtType: {
	    int vtype= namev[word].lptype & 0xffff & ~lpc_type;
	    *(namev[word].value) = vtype;
	    if (vtype==lpc_uni4 || vtype==lpc_br875) {
		/* set default values for these PIC programmers : */
#ifdef LPBDEBUG
	    fprintf(stderr," defaults chiptype=%d\n",vtype);
#endif
		p_power= parseBit("TxD");
		p_mclr= parseBit("TxD");
		p_data= parseBit("DTR");
		p_data_f= parseBit("CTS");
		p_clock= parseBit("RTS");
		p_clock_f= -1;
		p_no_power_C4C8 = (vtype==lpc_br875)? 1 : 0;
	    }
	    if (vtype==lpc_tait_invert) {
		p_power= parseBit("d2");
		p_mclr= parseBit("d3");
		p_data= parseBit("!d0");
		p_data_f= parseBit("!ack");
		p_clock= parseBit("!d1");
		p_clock_f= -1;
	    }
	    if (vtype==lpc_tait_direct) {
		p_power= parseBit("!d2");
		p_mclr= parseBit("!d3");
		p_data= parseBit("d0");
		p_data_f= parseBit("ack");
		p_clock= parseBit("d1");
		p_clock_f= -1;
	    }
	   }
	   break;
	}
    }/* end of for-each-line loop */
    fclose(fp);
    return(0);
}




static struct { char *name; int nameLen, value; } symBits[] = {
    { "busy",  4, BI_STAT | 7 },
    { "ack",   3, BI_STAT | 6 },
    { "error", 5, BI_STAT | 3 },

    { "TxD", 3, 030 | 6 },	/* base +3, bit 6 */
    { "RTS", 3, 040 | 1 },	/* base +4, bit 1 */
    { "DTR", 3, 040 | 0 },	/* base +4, bit 0 */
    { "CTS", 3, 060 | 4 },	/* base +6, bit 4 */

#ifndef OLDSYMTABLE
    { "d0",    2, BI_DATA | 0 },
    { "d1",    2, BI_DATA | 1 },
    { "d2",    2, BI_DATA | 2 },
    { "d3",    2, BI_DATA | 3 },
    { "d4",    2, BI_DATA | 4 },
    { "d5",    2, BI_DATA | 5 },
    { "d6",    2, BI_DATA | 6 },
    { "d7",    2, BI_DATA | 7 },
    { "c0",    2, BI_CTL  | 0 },
    { "c1",    2, BI_CTL  | 1 },
    { "c2",    2, BI_CTL  | 2 },
    { "c3",    2, BI_CTL  | 3 },
    { "c4",    2, BI_CTL  | 4 },
    { "c5",    2, BI_CTL  | 5 },
    { "c6",    2, BI_CTL  | 6 },
    { "c7",    2, BI_CTL  | 7 },
    { "s0",    2, BI_STAT | 0 },
    { "s1",    2, BI_STAT | 1 },
    { "s2",    2, BI_STAT | 2 },
    { "s3",    2, BI_STAT | 3 },
    { "s4",    2, BI_STAT | 4 },
    { "s5",    2, BI_STAT | 5 },
    { "s6",    2, BI_STAT | 6 },
    { "s7",    2, BI_STAT | 7 },
#endif
    { NULL,    0, 0 }
};

static int parseBit(char *spec)
{
    char *cp = spec;
    int val = 0,
    	valid = 0,
    	i;

#ifdef LPBDEBUG
    fprintf(stderr,"[parsing bit <%s>]\n", spec); 
#endif

    if (!*cp)
	return -1;

    do {
#ifdef LPBDEBUG
	fprintf(stderr,"@%s,%d\n", cp, val);  
#endif
	switch(*cp) {
	case ' ': case '\t': case '\n':
	    break;     /* ignore whitespace */

	case '!': case '~':
	    val ^= BI_INVERSE;  /* inversion */
	    break;

#ifdef OLDSYMTABLE
	case 's': case 'S':
	    /* s0 through s7  status-word bits 0 through 7. */
	    val &= ~(BI_TYPE | BI_OFFSET);
	    val |= strtol(cp+1, &cp, 10) | BI_STAT;
	    cp--;
	    valid = 1;
	    break;
#endif
	default:
#ifdef OLDSYMTABLE
	 if ( (*cp == 'd' || *cp == 'D' )
            && ( *(cp+1) >= '0' && *(cp+1) <='7' ) ) {
	    /* handle d0-d7, but accept DTR / DSR ... 
               d0 through d7  Data bits 0 through 7. */
	    val &= ~(BI_TYPE | BI_OFFSET);
	    val |= strtol(cp+1, &cp, 10) | BI_DATA;
	    cp--;
	    valid = 1;
	  } else if ( (*cp == 'c' || *cp == 'C' )
            && ( *(cp+1) >= '0' && *(cp+1) <='7' ) ) {
	    /* c0 through c7  control-word bits 0 through 7. */
	    val &= ~(BI_TYPE | BI_OFFSET);
	    val |= strtol(cp+1, &cp, 10) | BI_CTL;
	    cp--;
	    valid = 1;
	  } else 
#endif
	  {
	    /* The rest is searched in the symBits name table. */
	    for (i = 0; symBits[i].name; i++)
		if (!strncmp(cp, symBits[i].name, symBits[i].nameLen)) {
		    cp += symBits[i].nameLen - 1;
		    val &= ~(BI_TYPE | BI_OFFSET);
		    val |= symBits[i].value;
		    valid = 1;
		    break;
		}
	    if (symBits[i].name)
		break;

	    fprintf(stderr, "unrecognized pin name: \"%s\"\n", cp);
	    return -1;
	  }
	}
    } while(*++cp);
    if (val>=0) lpb.ports_in_use |= 1<<((val>>3) &0x7);
    // test:  fprintf(stderr," %0x\n",lpb.ports_in_use);
    return (valid ? val : -1);
}

static const char * const typeNames[8] = {
    "data", "status", "control", "txd", "lcr", "---", "msr", "---"
};

void
lpb_dump_names(int namec, struct lp_name *namev)
{
    int v, i;

    for (i=0; i<namec; i++) {
	fprintf(stderr,"%s = ", namev[i].name);
        if (namev[i].lptype==lpc_eof) continue;
	if ( namev[i].lptype==lpc_number) {	
	/* strcmp(namev[i].name, "base")) { */
	    fprintf(stderr,"0x%03x\n", *(namev[i].value));
	    continue;
	}
	v = *(namev[i].value);
	if (v == -1)
	    fprintf(stderr,"undefined\n");
	else
	    fprintf(stderr,"%s bit %d%s\n",
		typeNames[ (v & BI_TYPE) >> BI_TYPE_SHIFT ],
		v & BI_OFFSET,
		(v & BI_INVERSE) ? " inverted":"");
    }
}
